@@ -21,39 +21,25 @@ class Bash(object):
     def test(self, cmd="1", failure_message=""):
         """Equivalent to `bash -c '{init}; [[ {cmd} ]]'`."""
         init = self.init + "\n" if self.init else ""
-        proc = subprocess.Popen(
-            [
-                "bash",
-                "-o",
-                "pipefail",
-                "-ec",
-                "{init}[[ {cmd} ]]".format(init=init, cmd=cmd),
-            ]
-        )
+        proc = subprocess.Popen([
+            "bash", "-o", "pipefail", "-ec", "{init}[[ {cmd} ]]".format(init=init, cmd=cmd)])
         stdout, stderr = proc.communicate()
-        assert (
-            0 == proc.wait() and not stdout and not stderr
-        ), """\
+        assert (0 == proc.wait() and not stdout and not stderr), """\
 {}
 {}
 === stdout ===
 {}=== stderr ===
-{}""".format(
-            failure_message, cmd, stdout or "", stderr or ""
-        )
+{}""".format(failure_message, cmd, stdout or "", stderr or "")
 
     def compgen(self, compgen_cmd, word, expected_completions, failure_message=""):
         self.test(
-            '"$(echo $(compgen {} -- "{}"))" = "{}"'.format(
-                compgen_cmd, word, expected_completions
-            ),
+            '"$(echo $(compgen {} -- "{}"))" = "{}"'.format(compgen_cmd, word,
+                                                            expected_completions),
             failure_message,
         )
 
 
-@pytest.mark.parametrize(
-    "init,test", [("export FOO=1", '"$FOO" -eq 1'), ("", '-z "$FOO"')]
-)
+@pytest.mark.parametrize("init,test", [("export FOO=1", '"$FOO" -eq 1'), ("", '-z "$FOO"')])
 def test_bash(init, test):
     shell = Bash(init)
     shell.test(test)
@@ -105,9 +91,10 @@ def test_prog_scripts(shell, caplog, capsys):
         assert script_py == ["complete -o filenames -F _shtab_shtab script.py"]
     elif shell == "zsh":
         assert script_py == [
-            "#compdef script.py",
-            "_describe 'script.py commands' _commands",
-        ]
+            "#compdef script.py", "_describe 'script.py commands' _commands",
+            "'*::: :->script.py'", "script.py)"]
+    elif shell == "tcsh":
+        assert script_py == ["complete script.py \\"]
     else:
         raise NotImplementedError(shell)
 
@@ -210,17 +197,35 @@ def test_subparser_aliases(shell, caplog):
     if shell == "bash":
         shell = Bash(completion)
         shell.compgen('-W "${_shtab_test_subparsers[*]}"', "s", "sub")
-        shell.compgen('-W "$_shtab_test_pos_0_choices"', "s", "sub")
+        shell.compgen('-W "${_shtab_test_pos_0_choices[*]}"', "s", "sub")
         shell.compgen('-W "${_shtab_test_subparsers[*]}"', "x", "xsub")
-        shell.compgen('-W "$_shtab_test_pos_0_choices"', "x", "xsub")
+        shell.compgen('-W "${_shtab_test_pos_0_choices[*]}"', "x", "xsub")
         shell.compgen('-W "${_shtab_test_subparsers[*]}"', "y", "ysub")
-        shell.compgen('-W "$_shtab_test_pos_0_choices"', "y", "ysub")
+        shell.compgen('-W "${_shtab_test_pos_0_choices[*]}"', "y", "ysub")
         shell.test('"$($_shtab_test_sub_pos_0_COMPGEN o)" = "one"')
         shell.test('-z "$_shtab_test_COMPGEN"')
 
     assert not caplog.record_tuples
 
 
+@fix_shell
+def test_subparser_colons(shell, caplog):
+    parser = ArgumentParser(prog="test")
+    subparsers = parser.add_subparsers()
+    subparsers.add_parser("sub:cmd", help="help message")
+    with caplog.at_level(logging.INFO):
+        completion = shtab.complete(parser, shell=shell)
+    print(completion)
+
+    if shell == "bash":
+        shell = Bash(completion)
+        shell.compgen('-W "${_shtab_test_subparsers[*]}"', "s", "sub:cmd")
+        shell.compgen('-W "${_shtab_test_pos_0_choices[*]}"', "s", "sub:cmd")
+        shell.test('-z "$_shtab_test_COMPGEN"')
+
+    assert not caplog.record_tuples
+
+
 @fix_shell
 def test_add_argument_to_optional(shell, caplog):
     parser = ArgumentParser(prog="test")
@@ -249,7 +254,7 @@ def test_add_argument_to_positional(shell, caplog, capsys):
         with pytest.raises(SystemExit) as exc:
             sub._actions[-1](sub, Namespace(), shell)
             assert exc.type == SystemExit
-            assert exc.vaue.code == 0
+            assert exc.value.code == 0
     completion, err = capsys.readouterr()
     print(completion)
     assert completion_manual == completion.rstrip()
@@ -258,9 +263,9 @@ def test_add_argument_to_positional(shell, caplog, capsys):
     if shell == "bash":
         shell = Bash(completion)
         shell.compgen('-W "${_shtab_test_subparsers[*]}"', "c", "completion")
-        shell.compgen('-W "$_shtab_test_pos_0_choices"', "c", "completion")
-        shell.compgen('-W "$_shtab_test_completion_pos_0_choices"', "ba", "bash")
-        shell.compgen('-W "$_shtab_test_completion_pos_0_choices"', "z", "zsh")
+        shell.compgen('-W "${_shtab_test_pos_0_choices[*]}"', "c", "completion")
+        shell.compgen('-W "${_shtab_test_completion_pos_0_choices[*]}"', "ba", "bash")
+        shell.compgen('-W "${_shtab_test_completion_pos_0_choices[*]}"', "z", "zsh")
 
     assert not caplog.record_tuples
 
